/***
*		NewEcho.c -- a new & improved Echo Tool
*		 1989, 1990 The Blue Ribbon SoundWorks, Ltd.  &
*		 1995, 1996, 1997 Another Option   All rights reserved.
***/

/*
=============================================================================
|   DATE   | WHAT
=============================================================================
 12 Jan 96 | created
 14 Jan 96 | debugged & added random stuff
 15 Jan 96 | added pop-up slope gadget
 13 Apr 96 | added padedit
 15 Nov 96 | regargs & tweeked; added snap/sysex
  2 Feb 97 | fixed missing noteoffs, apparently from the previous tweeks. really
           | added sysex. what was wrong with you?
 27 Oct 97 | real-time controllers ui
 28 Oct 97 | ditto, only more.
 29 Oct 97 | controllers real stuff -- they don't work from the toolpad; this tool
           | won't work for events without durations!!
 17 Nov 97 | added 0 repeats; it does work with no durations, but will get screwy 
           | if the destinations change between noteon and noteoff
  3 Mar 98 | sysex was not passed to the next tool, dillhole.
*/

#define DBUG 0
//#define FILEDUMP 1

#include <exec/types.h>
#include <exec/memory.h>
#include <dos/dos.h>
#include <intuition/intuition.h>
#include <proto/exec.h>
#include <proto/graphics.h>
#include <proto/intuition.h>
#include <clib/alib_protos.h>
#include <clib/alib_stdio_protos.h>
//#include <stdio.h>
#include <string.h>

#include "bars.h"
#include "edit.h"
#include "barsMacros.h"
#include "bplib.h"
#include "configlib.h"
#include "popuppipes.h"
#include "NewEchow.h"
#include "NewEchoPop.h"
#include "/icons/NewEchoDown.h"
#include "/icons/NewEchoUp.h"
#include "ToolID.h"
#include "snapshotgadget.h"
#include "configgadget2.h"
#include "proto/newecho_protos.h"

#include "debug.h"

/* maximum number of echoes */
#define MAX_REPEATS 30
/* maximum change in byte2 between echoes */
#define MAX_CHANGE  25
/* upper limit for the time between echoes */
#define MAX_DELAY 1536

static UBYTE const versiontag[] = "$VER: NewEcho 40.42 (03.03.98)";

/* refresh tool window on sysex */
#define SECRET_MESSAGE 9999
/* updated via cc message */
#define SECRET_CC (SECRET_MESSAGE - 1)

#define CTRL_COUNT 7

struct EchoTool {
	struct Tool tool;
	short repeat; // was echocount;
	short delay;  // was echodistance;
	char echostyle;
	/* new goodies */
	UWORD flags;
	char change;
	char controller; // 121==all
	char min_repeat;
	char max_repeat;
	char min_change;
	char max_change;
	struct ControllerInfo cinfo[CTRL_COUNT];
};

#define SYSEX_START tool->repeat
#define SX_LENGTH (sizeof(struct EchoTool)-sizeof(struct Tool))

/*** tool flags ***/
/* events to echo */
#define NOTES              1
#define PBEND              2
#define MTOUCH             4
#define PTOUCH             8
#define CCHG            0x10
/* behavioral stuff */
#define TOOL_ON         0x20
#define BRANCH_SOURCE   0x40
#define BRANCH_ECHOES   0x80
#define USE_SYSEX      0x100
#define FREE_SYSEX     0x200
#define USE_CONTROLS   0x400
#define SHOW_CCNUMS    0x800
#define DELETE_SOURCE 0x1000

/*** echo styles ***/
#define STYLE_FALLING 0
#define STYLE_LEVEL   1
#define STYLE_RISING  2
#define STYLE_RANDOM  3

/*** defaults ***/
#define DEFAULT_FLAGS (NOTES | TOOL_ON | FREE_SYSEX | SHOW_CCNUMS)
/* this is really a change of 10 because of allowing random amounts */
#define DEFAULT_CHANGE 11

#define tw (tool->tool.window)

extern struct IntuitionBase *IntuitionBase;
extern struct GfxBase *GfxBase;
extern struct LayersBase *LayersBase;
extern struct DOSBase *DOSBase;
extern struct ExecBase *SysBase;
extern struct Functions *functions;

/********************************************************************/
static void __regargs inittool(struct EchoTool *tool)
{
	tool->repeat = 4;
	tool->delay = 96;
	tool->echostyle = STYLE_FALLING;
	tool->flags = DEFAULT_FLAGS;
	tool->change = DEFAULT_CHANGE;
	tool->controller = 121;
	tool->min_repeat = 1;
	tool->max_repeat = 6; //really 5 because of allowing random
	tool->min_change = 1;
	tool->max_change = 10;
	InitControlInfoArray(&tool->cinfo[0],CTRL_COUNT,CITYPE_NODIFF|CITYPE_OFF,
	                     CIFLAG_DISCARD);
	tool->tool.touched = TOUCH_INIT;
}

#pragma msg 104 ignore push

/********************************************************************/
static long __stdargs __saveds hiloRnd(struct Window *w,struct Gadget *g,long h)

{
   char hiloText[7];

   SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset,
            (g->Flags & GADGDISABLED) ? HILO_DPEN : HILO_APEN, HILO_BPEN,
				HILO_DRAWMODE);
	if (h > 0) {
//		sprintf(hiloText,"%2ld    ",(g->GadgetID==INCREMENT) ? h-1 : h);
		sprintf(hiloText,"%2ld    ",h-1);
	}
	else {
		sprintf(hiloText,"%s","Random");
	}
	Text(w->RPort,hiloText,6);
	return(h);
}

/********************************************************************/
static long __stdargs __saveds hilo2(struct Window *w,struct Gadget *g,long h)

{
   char hiloText[3];

   SetRPort(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset,
            (g->Flags & GADGDISABLED) ? HILO_DPEN : HILO_APEN, HILO_BPEN,
				HILO_DRAWMODE);
	sprintf(hiloText,"%2ld",h);
	Text(w->RPort,hiloText,2);
	return(h);
}

#pragma msg 104 pop

struct EchoTool *selectedtool = NULL;

/***
 *** displays time mm.bb.ccc, with a minimum of 1, for prop gadget 
 ***/

static long __stdargs __saveds hiloShort(struct Window *w,struct Gadget *g,long  x)
{
	char text[13];

	if (x==0) x=1;
	functions->lengthtostring(selectedtool->tool.clip,0,x,text);
	SetRPort(w->RPort, g->LeftEdge+g->Width+XOffset, g->TopEdge+YOffset,
            (g->Flags & GADGDISABLED)?HILO_DPEN:HILO_APEN,
				HILO_BPEN, HILO_DRAWMODE);
	Text(w->RPort,&text[3],8);
	return((long)x);
}

static char * const ccnames[] = {
/*0 - 4 */	"MonPres", "ModWhel", "Breath ", "MIDI 03", "Foot   ",
/*5 - 9 */	"PortTim", "DataEnt", "Volume ", "Balance", "MIDI 09",
/*10-14 */	"Pan    ", "Express", "MIDI 12", "MIDI 13", "MIDI 14",
/*15-19 */	"MIDI 15", "Genl 1 ", "Genl 2 ", "Genl 3 ", "Genl 4 ",
/*20-24 */	"MIDI 20", "MIDI 21", "MIDI 22", "MIDI 23", "MIDI 24",
/*25-29 */	"MIDI 25", "MIDI 26", "MIDI 27", "MIDI 28", "MIDI 29",
/*30-34 */	"MIDI 30", "MIDI 31", "MIDI 32", "MIDI 33", "MIDI 34",
/*35-39 */	"MIDI 35", "MIDI 36", "MIDI 37", "MIDI 38", "MIDI 39",
/*40-44 */	"MIDI 40", "MIDI 41", "MIDI 42", "MIDI 43", "MIDI 44",
/*45-49 */	"MIDI 45", "MIDI 46", "MIDI 47", "MIDI 48", "MIDI 49",
/*50-54 */	"MIDI 50", "MIDI 51", "MIDI 52", "MIDI 53", "MIDI 54",
/*55-59 */	"MIDI 55", "MIDI 56", "MIDI 57", "MIDI 58", "MIDI 59",
/*60-64 */	"MIDI 60", "MIDI 61", "MIDI 62", "MIDI 63", "Sustain",
/*65-69 */	"PortSwt", "Sosten.", "SoftPed", "MIDI 68", "FrezPdl",
/*70-74 */	"MIDI 70", "MIDI 71", "MIDI 72", "MIDI 73", "MIDI 74",
/*75-79 */	"LegtoSw", "MIDI 76", "MIDI 77", "MIDI 78", "MIDI 79",
/*80-84 */	"Genl 5 ", "Genl 6 ", "Genl 7 ", "Genl 8 ", "MIDI 84",
/*85-89 */	"MIDI 85", "MIDI 86", "MIDI 87", "MIDI 88", "MIDI 89",
/*90-94 */	"MIDI 90", "FXDepth", "TremDep", "ChorDep", "CelDep ",
/*95-99 */	"PhsrDep", "DataInc", "DataDec", "MIDI 98", "MIDI 99",
/*100-104*/	"MIDI100", "MIDI101", "MIDI102", "MIDI103", "MIDI104",
/*105-109*/	"MIDI105", "MIDI106", "MIDI107", "MIDI108", "MIDI109",
/*110-114*/	"MIDI110", "MIDI111", "MIDI112", "MIDI113", "MIDI114",
/*115-119*/	"MIDI115", "MIDI116", "MIDI117", "MIDI118", "MIDI119",
/*120-124*/	"MIDI120" /*, "RsetCtl", "LoclCtl", "AllOff ", "OmniOff",*/
/*125-127*/	/*"Omni On", "Mono On", "Poly On"*/
};

/*** display pot value where 121 == All ***/
static long __stdargs __saveds hiloAll(struct Window *w, struct Gadget *g, long h)

{
	char text[8];
	
	SetRPort(w->RPort, g->LeftEdge+g->Width+XOffset, g->TopEdge+YOffset,
            (g->Flags & GADGDISABLED)?HILO_DPEN:HILO_APEN,
				HILO_BPEN, HILO_DRAWMODE);
	if (h < 121) {
		if (selectedtool->flags & SHOW_CCNUMS) {
#pragma msg 104 ignore push
			sprintf(text,"%3ld    ",h);
#pragma msg 104 ignore pop
//			strcat(&text[3],"    ");
//			Text(w->RPort,text,7);
		}
		else {
			//Text(w->RPort,ccnames[h],7);
			strcpy(text,ccnames[h]);
		}
	}
	else {
		sprintf(text,"%s","All    ");
	}
	Text(w->RPort,text,7);
	return(h);
}

/* pens for config pot value text */
#define TEXTCOLOR 1
#define BACKCOLOR 0

/* display pot value where 0 is Off */

static long __stdargs __saveds hiloOff(struct Window *w, struct Gadget *g, long h)

{
	char text[8];
	
	Move(w->RPort,g->LeftEdge+g->Width+XOffset,g->TopEdge+YOffset);
	SetAPen(w->RPort,TEXTCOLOR);
	SetBPen(w->RPort,BACKCOLOR);
	SetDrMd(w->RPort,JAM2);
	if (--h > -1) {
		if (selectedtool->flags & SHOW_CCNUMS) {
#pragma msg 104 ignore push
			sprintf(text,"%3ld    ",h);
#pragma msg 104 ignore pop
//			strcat(&text[3],"    ");
			Text(w->RPort,text,7);
		}
		else {
			Text(w->RPort,ccnames[h],7);
		}
	}
	else {
#pragma msg 104 ignore push
		sprintf(text,"%s","Off    ");
#pragma msg 104 ignore pop
//		strcat(&text[3],"    ");
		Text(w->RPort,text,7);
	}
	return(h);
}

/*==========================================================================
* dosnapshot
*
*  INPUT: pointer to tool
* RETURN: none
*
* EFFECT: makes a sysex message and posts it
*=========================================================================*/

static void dosnapshot(struct EchoTool *tool)

{
	UBYTE sx[SX_LENGTH];
	struct StringEvent *event;
	
	memcpy(sx,&SYSEX_START,SX_LENGTH);
	if (event = BuildSnapshot(&tool->tool, 
	                          functions->running ? functions->timenow : -1,
	                          sx, SX_LENGTH)) {
		PostSnapshot(&tool->tool,event);
	}
}

/**********************************************/
/* twiddle and redraw config discard gadgets  */

static void __regargs refreshConfigGadgets(struct EchoTool *tool)
{
	long i;
	struct Gadget *gadget;

	for (i=C_STATE ; i<=C_CHANGE ; i++) {
		gadget = functions->GetGadget(tw,i+CTRL_COUNT); //respective discard gadget
		if (tool->cinfo[i-C_STATE].flags & CIFLAG_DISCARD) {
			gadget->Flags |= SELECTED;
		}
		else {
			gadget->Flags &= ~SELECTED;
		}
		functions->RefreshGadget(tool->tool.window,i+CTRL_COUNT);
	}
}

/*************************************/
/* refresh real-time controller pots */

static void __regargs updateConfigPots(struct EchoTool *tool)
{
	UBYTE i;

	selectedtool = tool;
	for (i=C_STATE ; i<=C_CHANGE ; i++) {
		functions->ModifyEmbossedProp(tw,i,tool->cinfo[i-C_STATE].controller+1,
			                           0,0,0,0,0);
		functions->DrawEmbossedProp(tw,i);
		functions->DrawEmbossed(tw,i+GUP);
		functions->DrawEmbossed(tw,i+GDOWN);
	}
}

#define REFRESH_BOOL        1
#define REFRESH_STYLE       2
#define REFRESH_INCREMENT   4
#define REFRESH_REPEAT      8
#define REFRESH_DELAY      16
#define REFRESH_CONFIG     32
#define REFRESH_DRAWCONFIG 64
#define REFRESH_POPPIPES  128

#define popmenu &PopMenu

/********************************************************************/
static void __stdargs __saveds edittool(struct EchoTool *tool)
{
	struct IntuiMessage *message;
	long class, code;
	struct Gadget *gadget;
	struct NewWindow *nw;
	struct Menu *menu;
	struct MenuItem *menuitem;
//	struct Image *image;
	UBYTE refresh;

	if (!tool->tool.touched) inittool(tool);
/*??? why the hell is this backwards ??*/
	if (InToolPad(&tool->tool)) refresh = 40;
	else refresh = 0;
	nw = OpenEmbossedWindowSlop(&NewWindowStructure1,&tool->tool,0,WTITLE,refresh);
	if (!nw) return;
	if (tool->flags & USE_SYSEX) MenuItem7.Flags |= CHECKED;
	else MenuItem7.Flags &= ~CHECKED;
	if (tool->flags & FREE_SYSEX) MenuItem8.Flags |= CHECKED;
	else MenuItem8.Flags &= ~CHECKED;
	if (tool->flags & USE_CONTROLS) MenuItem10.Flags |= CHECKED;
	else MenuItem10.Flags &= ~CHECKED;
	if (tool->flags & SHOW_CCNUMS) MenuItem11.Flags |= CHECKED;
	else MenuItem11.Flags &= ~CHECKED;
	if (refresh) {
		MenuItem8.Flags &= ~ITEMENABLED; // no freeing sysex in toolpad
		MenuItem9.Flags &= ~ITEMENABLED; // no snapshots either
	}
	else {
		MenuItem8.Flags |= ITEMENABLED;
		MenuItem9.Flags |= ITEMENABLED;
	}
	if (menu = functions->DupeMenu(&MenuList1)) SetMenuStrip(tw,menu);
	else SetWindowTitles(tw,(UBYTE *)-1,MenuList1.MenuName);
	PrintIText(tw->RPort,&IntuiTextList1,0,0);
	EmbossOnGList(tw,&Gadget1,GCOUNT,1);
	functions->FatEmbossedPropOn(tw,DELAY,DELAY+GDOWN,DELAY+GUP,hiloShort,
	                             MAX_DELAY+1,1);
	functions->FatEmbossedPropOn(tw,REPEAT,REPEAT+GDOWN,REPEAT+GUP,hiloRnd,
	                             MAX_REPEATS+1,1);
	functions->FatEmbossedPropOn(tw,CONTROLLER,CONTROLLER+GDOWN,CONTROLLER+GUP,hiloAll,122,1);
	functions->FatEmbossedPropOn(tw,INCREMENT,INCREMENT+GDOWN,INCREMENT+GUP,hiloRnd,
	                             MAX_CHANGE+2,1);
	functions->FatEmbossedPropOn(tw,REPEAT_MIN,REPEAT_MIN+GDOWN,REPEAT_MIN+GUP,hilo2,
	                             MAX_REPEATS+1,1);
	functions->FatEmbossedPropOn(tw,REPEAT_MAX,REPEAT_MAX+GDOWN,REPEAT_MAX+GUP,hilo2,
	                             MAX_REPEATS+1,1);
	functions->FatEmbossedPropOn(tw,INCREMENT_MIN,INCREMENT_MIN+GDOWN,INCREMENT_MIN+GUP,
	                             hilo2,MAX_CHANGE+1,1);
	functions->FatEmbossedPropOn(tw,INCREMENT_MAX,INCREMENT_MAX+GDOWN,INCREMENT_MAX+GUP,
	                             hilo2,MAX_CHANGE+1,1);
	for (code = C_STATE ; code <= C_CHANGE ; code++ ) { //emboss config gadgets
		functions->FatEmbossedPropOn(tw,code,code+GDOWN,code+GUP,hiloOff,122,1);
	}
	if (refresh) {              // these don't work from the toolpad
		functions->EnableEmbossed(tw,ON,0);
		functions->EnableEmbossed(tw,OFF,0);
	}
	else {
		AddConfigGadget(tw);
		AddSnapshotGadget(tw); // and this doesn't either
	}
	if (WindowIsTall(tw)) {
		DrawConfigFrame(tw);
		refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
	}
	else refresh = 0;

	refresh |= REFRESH_STYLE|REFRESH_BOOL|REFRESH_DELAY|REFRESH_INCREMENT|REFRESH_REPEAT|REFRESH_POPPIPES;

	for ( ; ; ) {
		if (refresh) {
			if (refresh & REFRESH_STYLE) {
				functions->SelectEmbossed(tw,FALLING,tool->echostyle == STYLE_FALLING);
				functions->SelectEmbossed(tw,LEVEL,tool->echostyle == STYLE_LEVEL);
				functions->SelectEmbossed(tw,RISING,tool->echostyle == STYLE_RISING);
				functions->SelectEmbossed(tw,RANDOM,tool->echostyle == STYLE_RANDOM);
/*
				if (tool->echostyle == STYLE_FALLING) image = &FallingImage;
				else if (tool->echostyle == STYLE_LEVEL) image = &LevelImage;
				else if (tool->echostyle == STYLE_RISING) image = &RisingImage;
				else if (tool->echostyle == STYLE_RANDOM) image = &RandomImage;
				DrawImage(tw->RPort,image,POPIMAGE_X,POPIMAGE_Y);
*/
				refresh |= REFRESH_INCREMENT;
			}

			if (refresh & REFRESH_BOOL) {
				functions->SelectEmbossed(tw,ON,tool->flags & TOOL_ON);
				functions->SelectEmbossed(tw,OFF,!(tool->flags & TOOL_ON));
				functions->SelectEmbossed(tw,NOTESG,tool->flags & NOTES);
				functions->SelectEmbossed(tw,BENDG,tool->flags & PBEND);
				functions->SelectEmbossed(tw,MTOUCHG,tool->flags & MTOUCH);
				functions->SelectEmbossed(tw,PTOUCHG,tool->flags & PTOUCH);
				functions->SelectEmbossed(tw,CCHGG,tool->flags & CCHG);
				functions->SelectEmbossed(tw,SOURCE,tool->flags & BRANCH_SOURCE);
				functions->SelectEmbossed(tw,ECHOES,tool->flags & BRANCH_ECHOES);
				selectedtool = tool;
				EnableFatProp(tw,CONTROLLER,CONTROLLER+GUP,CONTROLLER+GDOWN,
				              code=(tool->flags & CCHG));
				if (code) {
					functions->ModifyEmbossedProp(tw,CONTROLLER,tool->controller,0,0,0,0,0);
				}
				functions->DrawEmbossedProp(tw,CONTROLLER);
			}

			selectedtool = tool;
			if (refresh & REFRESH_DELAY) {
				if (tool->delay > MAX_DELAY) tool->delay = MAX_DELAY;
				functions->ModifyEmbossedProp(tw,DELAY,tool->delay,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,DELAY);
			}
			if (refresh & REFRESH_INCREMENT) {
				EnableFatProp(tw,INCREMENT,INCREMENT+GUP,INCREMENT+GDOWN,
				              code = (tool->echostyle != STYLE_LEVEL));
				if (code) {
					functions->ModifyEmbossedProp(tw,INCREMENT,tool->change,0,0,0,0,0);
				}
				functions->DrawEmbossedProp(tw,INCREMENT);
				EnableFatProp(tw,INCREMENT_MAX,INCREMENT_MAX+GUP,INCREMENT_MAX+GDOWN,
		             code=((tool->change == 0) && (tool->echostyle != STYLE_LEVEL)));
				EnableFatProp(tw,INCREMENT_MIN,INCREMENT_MIN+GUP,INCREMENT_MIN+GDOWN,code);
				if (code) {
					if (tool->max_change < tool->min_change) {
						tool->max_change = tool->min_change;
					}
					functions->ModifyEmbossedProp(tw,INCREMENT_MAX,tool->max_change,0,0,0,0,0);
					functions->ModifyEmbossedProp(tw,INCREMENT_MIN,tool->min_change,0,0,0,0,0);
				}
				functions->DrawEmbossedProp(tw,INCREMENT_MAX);
				functions->DrawEmbossedProp(tw,INCREMENT_MIN);
			}
				
			if (refresh & REFRESH_REPEAT) {
				functions->ModifyEmbossedProp(tw,REPEAT,tool->repeat,0,0,0,0,0);
				functions->DrawEmbossedProp(tw,REPEAT);
				EnableFatProp(tw,REPEAT_MAX,REPEAT_MAX+GUP,REPEAT_MAX+GDOWN,
				              code=(tool->repeat == 0));
				EnableFatProp(tw,REPEAT_MIN,REPEAT_MIN+GUP,REPEAT_MIN+GDOWN,code);
				if (code) {
					if (tool->max_repeat < tool->min_repeat) {
						tool->max_repeat = tool->min_repeat;
					}
					functions->ModifyEmbossedProp(tw,REPEAT_MAX,tool->max_repeat,0,0,0,0,0);
					functions->ModifyEmbossedProp(tw,REPEAT_MIN,tool->min_repeat,0,0,0,0,0);
				}
				functions->DrawEmbossedProp(tw,REPEAT_MAX);
				functions->DrawEmbossedProp(tw,REPEAT_MIN);
			}
			if (refresh & REFRESH_CONFIG) {
				UpdateControllers(&tool->cinfo[0]);
				updateConfigPots(tool);
			}
			if (refresh & REFRESH_DRAWCONFIG) {
				refreshConfigGadgets(tool);
			}
			if (refresh & REFRESH_POPPIPES) {
				gadget = functions->GetGadget(tw,ECHO_POP);
				if (tool->flags & BRANCH_ECHOES) {
					gadget->GadgetRender = (APTR)&ElbowImage;
				}
				else gadget->GadgetRender = (APTR)&PipeImage;
				gadget = functions->GetGadget(tw,SOURCE_POP);
				if (tool->flags & BRANCH_SOURCE) {
					gadget->GadgetRender = (APTR)&ElbowImage;
				}
				else if (tool->flags & DELETE_SOURCE) {
					gadget->GadgetRender = (APTR)&PlugImage;
				}
				else {
					gadget->GadgetRender = (APTR)&PipeImage;
				}
				RefreshGList(gadget,tw,NULL,2);
				refresh = 0;
			}
		}
		message = (struct IntuiMessage *)functions->GetIntuiMessage(tw);
		if (message->ExecMessage.mn_Length == SECRET_MESSAGE) {
			refresh = 0xFF;
			functions->myfree((char *)message,sizeof(struct Message));
			continue;
		}
		else if (message->ExecMessage.mn_Length == SECRET_CC) {
			refresh = REFRESH_STYLE|REFRESH_BOOL|REFRESH_DELAY|REFRESH_INCREMENT|REFRESH_REPEAT|REFRESH_POPPIPES;
			functions->myfree((char *)message,sizeof(struct Message));
			continue;
		}
		/* else */
		class = message->Class;
		code = message->Code;
		gadget = (struct Gadget *) message->IAddress;
		class = functions->SystemGadgets(tw,class,gadget,code);
		ReplyMsg((struct Message *)message);

		if (class == CLOSEWINDOW) break;

		else if (class == NEWSIZE) {
			if (WindowIsTall(tw)) {
				DrawConfigFrame(tw);
				refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
			}
			SizeTitleBar(tw,-40);
			functions->RefreshGadget(tw,SNAPSHOT_GADGET_ID);
			functions->RefreshGadget(tw,CONFIG_GADGET_ID);
		}

		else if (class == MENUPICK) {
			while (code != MENUNULL) {
				menuitem = ItemAddress(menu,code);
				switch (ITEMNUM(code)) {
					case 0 :
						tool->flags ^= USE_SYSEX;
						break;
					case 1 :
						tool->flags ^= FREE_SYSEX;
						break;
					case 2 :
						dosnapshot(tool);
						break;
					case 3 :
						tool->flags ^= USE_CONTROLS;
						break;
					case 4 :
						tool->flags ^= SHOW_CCNUMS;
						updateConfigPots(tool);
						functions->DrawEmbossedProp(tw,CONTROLLER);
//						refresh = REFRESH_BOOL;
						break;
				}
				code = menuitem->NextSelect;
			}
		}

		else if (class == GADGETDOWN) {
			selectedtool = tool;
			switch (gadget->GadgetID) {
				case SOURCE_POP :
					PlugMenuItem.Flags |= ITEMENABLED;
					code = (short)functions->PopUpMenu(tw,&PopPipeMenu);
					if (ITEMNUM(code) < 3) {
						switch (ITEMNUM(code)) {
							case ELBOW_ITEM :
								tool->flags = (tool->flags & ~DELETE_SOURCE) | BRANCH_SOURCE;
								break;
							case PLUG_ITEM :
								tool->flags = (tool->flags & ~BRANCH_SOURCE) | DELETE_SOURCE;
								break;
							case PIPE_ITEM :
								tool->flags &= ~(BRANCH_SOURCE | DELETE_SOURCE);
								break;
						}
						refresh = REFRESH_POPPIPES;
					}
					break;
				case ECHO_POP :
					PlugMenuItem.Flags &= ~ITEMENABLED;
					code = (short)functions->PopUpMenu(tw,&PopPipeMenu);
					if (ITEMNUM(code) < 3) {
						switch (ITEMNUM(code)) {
							case ELBOW_ITEM :
								tool->flags |= BRANCH_ECHOES;
								break;
							case PIPE_ITEM :
								tool->flags &= ~BRANCH_ECHOES;
								break;
						}
						refresh = REFRESH_POPPIPES;
					}
					break;
				case INCREMENT :
					tool->change = functions->DragEmbossedProp(tw,INCREMENT);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT+GDOWN :
					tool->change = functions->ShiftEmbossedProp(tw,INCREMENT,-1,0);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT+GUP :
					tool->change = functions->ShiftEmbossedProp(tw,INCREMENT,1,0);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MAX :
					tool->max_change = functions->DragEmbossedProp(tw,INCREMENT_MAX);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MAX+GDOWN :
					tool->max_change = functions->ShiftEmbossedProp(tw,INCREMENT_MAX,-1,0);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MAX+GUP :
					tool->max_change = functions->ShiftEmbossedProp(tw,INCREMENT_MAX,1,0);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MIN :
					tool->min_change = functions->DragEmbossedProp(tw,INCREMENT_MIN);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MIN+GDOWN :
					tool->min_change = functions->ShiftEmbossedProp(tw,INCREMENT_MIN,-1,0);
					refresh = REFRESH_INCREMENT;
					break;
				case INCREMENT_MIN+GUP :
					tool->min_change = functions->ShiftEmbossedProp(tw,INCREMENT_MIN,1,0);
					refresh = REFRESH_INCREMENT;
					break;

				case REPEAT :
					tool->repeat = functions->DragEmbossedProp(tw,REPEAT);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT+GDOWN :
					tool->repeat = functions->ShiftEmbossedProp(tw,REPEAT,-1,0);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT+GUP :
					tool->repeat = functions->ShiftEmbossedProp(tw,REPEAT,1,0);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MAX :
					tool->max_repeat = functions->DragEmbossedProp(tw,REPEAT_MAX);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MAX+GDOWN :
					tool->max_repeat = functions->ShiftEmbossedProp(tw,REPEAT_MAX,-1,0);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MAX+GUP :
					tool->max_repeat = functions->ShiftEmbossedProp(tw,REPEAT_MAX,1,0);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MIN :
					tool->min_repeat = functions->DragEmbossedProp(tw,REPEAT_MIN);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MIN+GDOWN :
					tool->min_repeat = functions->ShiftEmbossedProp(tw,REPEAT_MIN,-1,0);
					refresh = REFRESH_REPEAT;
					break;
				case REPEAT_MIN+GUP :
					tool->min_repeat = functions->ShiftEmbossedProp(tw,REPEAT_MIN,1,0);
					refresh = REFRESH_REPEAT;
					break;

				case CONTROLLER :
					tool->controller = functions->DragEmbossedProp(tw,CONTROLLER);
					refresh = 1;
					break;
				case CONTROLLER+GDOWN :
					tool->controller = functions->ShiftEmbossedProp(tw,CONTROLLER,-1,0);
					refresh = 1;
					break;
				case CONTROLLER+GUP :
					tool->controller = functions->ShiftEmbossedProp(tw,CONTROLLER,1,0);
					refresh = 1;
					break;

				case DELAY :
					tool->delay = functions->DragEmbossedProp(tw,DELAY);
					refresh= REFRESH_DELAY;
					break;
				case DELAY+GDOWN :
					tool->delay = functions->ShiftEmbossedProp(tw,DELAY,-1,0);
					refresh= REFRESH_DELAY;
					break;
				case DELAY+GUP :
					tool->delay = functions->ShiftEmbossedProp(tw,DELAY,1,0);
					refresh= REFRESH_DELAY;
					break;
				case TRIPLET :
					tool->delay = (tool->delay << 1) / 3;
					refresh= REFRESH_DELAY;
					break;
				case DOT :
					tool->delay += (tool->delay >> 1);
					refresh= REFRESH_DELAY;
					break;
				case X2 :
					tool->delay <<= 1;
					refresh= REFRESH_DELAY;
					break;
				case DELAY_POP :
					code = functions->popupnote(tw,-1);
					if (code >= 0) {
						tool->delay = 768 >> code;
					}
					refresh= REFRESH_DELAY;
					break;
/*
				case SLOPE :
					code = (short)functions->PopUpMenu(tw,popmenu);
					if (ITEMNUM(code) < 4) {
						tool->echostyle = ITEMNUM(code);
						refresh = REFRESH_STYLE;
					}
					break;
*/
				case C_STATE:
				case C_DELAY:
				case C_ECHOES:
				case C_SLOPE:
				case C_BRANCHE:
				case C_BRANCHS:
				case C_CHANGE:
					tool->cinfo[gadget->GadgetID - C_STATE].controller =
						functions->DragEmbossedProp(tw,gadget->GadgetID);
					refresh = REFRESH_CONFIG;
					break;
				case C_STATE+GDOWN :
				case C_DELAY+GDOWN :
				case C_ECHOES+GDOWN :
				case C_SLOPE+GDOWN :
				case C_BRANCHE+GDOWN :
				case C_BRANCHS+GDOWN :
				case C_CHANGE+GDOWN :
					tool->cinfo[gadget->GadgetID - C_STATE - GDOWN].controller =
						   functions->ShiftEmbossedProp(tw,gadget->GadgetID-GDOWN,-1,0);
					refresh = REFRESH_CONFIG;
					break;
				case C_STATE+GUP :
				case C_DELAY+GUP :
				case C_ECHOES+GUP :
				case C_SLOPE+GUP :
				case C_BRANCHE+GUP :
				case C_BRANCHS+GUP :
				case C_CHANGE+GUP :
					tool->cinfo[gadget->GadgetID - C_STATE - GUP].controller =
						   functions->ShiftEmbossedProp(tw,gadget->GadgetID-GUP,1,0);
					refresh = REFRESH_CONFIG;
					break;
			}
		}

		else if (class == GADGETUP) {
			switch (gadget->GadgetID) {
				case SNAPSHOT_GADGET_ID :
					dosnapshot(tool);
					break;
				case CONFIG_GADGET_ID :
					ZoomWindow(tw);
					break;

				case FALLING :
					tool->echostyle = STYLE_FALLING;
					refresh = REFRESH_STYLE;
					break;
				case LEVEL :
					tool->echostyle = STYLE_LEVEL;
					refresh = REFRESH_STYLE;
					break;
				case RISING :
					tool->echostyle = STYLE_RISING;
					refresh = REFRESH_STYLE;
					break;
				case RANDOM :
					tool->echostyle = STYLE_RANDOM;
					refresh = REFRESH_STYLE;
					break;
				
				case DIS_STATE:
				case DIS_DELAY:
				case DIS_ECHOES:
				case DIS_SLOPE:
				case DIS_BRANCHE:
				case DIS_BRANCHS:
				case DIS_CHANGE:
					tool->cinfo[gadget->GadgetID - DIS_STATE].flags ^= CIFLAG_DISCARD;
					refresh = REFRESH_CONFIG | REFRESH_DRAWCONFIG;
					break;
				case ON :
					tool->flags |= TOOL_ON;
					refresh = REFRESH_BOOL;
					break;
				case OFF :
					tool->flags &= ~TOOL_ON;
					refresh = REFRESH_BOOL;
					break;
				case NOTESG :
					tool->flags ^= NOTES;
					refresh = REFRESH_BOOL;
					break;
				case BENDG :
					tool->flags ^= PBEND;
					refresh = REFRESH_BOOL;
					break;
				case MTOUCHG :
					tool->flags ^= MTOUCH;
					refresh = REFRESH_BOOL;
					break;
				case PTOUCHG :
					tool->flags ^= PTOUCH;
					refresh = REFRESH_BOOL;
					break;
				case CCHGG :
					tool->flags ^= CCHG;
					refresh = REFRESH_BOOL;
					break;
				case SOURCE :
					tool->flags ^= BRANCH_SOURCE;
//					if (tool->flags & BRANCH_SOURCE) tool->flags &= ~BRANCH_ECHOES;
					refresh = REFRESH_BOOL;
					break;
				case ECHOES :
					tool->flags ^= BRANCH_ECHOES;
//					if (tool->flags & BRANCH_ECHOES) tool->flags &= ~BRANCH_SOURCE;
					refresh = REFRESH_BOOL;
					break;
			}
		}
	}
	selectedtool = NULL;
	EmbossOffGList(tw,&GadgetList1,GCOUNT);
	functions->FatEmbossedPropOff(tw,DELAY,DELAY+GUP,DELAY+GDOWN);
	functions->FatEmbossedPropOff(tw,REPEAT,REPEAT+GUP,REPEAT+GDOWN);
	functions->FatEmbossedPropOff(tw,REPEAT_MAX,REPEAT_MAX+GUP,REPEAT_MAX+GDOWN);
	functions->FatEmbossedPropOff(tw,REPEAT_MIN,REPEAT_MIN+GUP,REPEAT_MIN+GDOWN);
	functions->FatEmbossedPropOff(tw,CONTROLLER,CONTROLLER+GUP,CONTROLLER+GDOWN);
	functions->FatEmbossedPropOff(tw,INCREMENT,INCREMENT+GUP,INCREMENT+GDOWN);
	functions->FatEmbossedPropOff(tw,INCREMENT_MAX,INCREMENT_MAX+GUP,INCREMENT_MAX+GDOWN);
	functions->FatEmbossedPropOff(tw,INCREMENT_MIN,INCREMENT_MIN+GUP,INCREMENT_MIN+GDOWN);
	for (code = C_STATE ; code <= C_CHANGE ; code++ ) {
		functions->FatEmbossedPropOff(tw,code,code+GDOWN,code+GUP);
	}
	NewCloseEmbossedWindow(&tool->tool,nw);
	if (menu) functions->DeleteMenu(menu);
	return;
}
/****************************************************************/
/* initializes stuff for an echo */

static void __regargs setup(struct EchoTool *tool, struct NoteEvent *event,
                            short *count, short *inc)
{
	char style;
	char echochange;

	if (tool->repeat) *count = tool->repeat-1;
	else *count = functions->random(tool->max_repeat-tool->min_repeat)+tool->min_repeat;

	if (tool->change) echochange = tool->change-1;
	else echochange = functions->random(tool->max_change-tool->min_change)+tool->min_change;

	if (tool->echostyle == STYLE_RANDOM) style = functions->random(65535) % 3;
	else style = tool->echostyle;
	if (style == STYLE_FALLING) *inc = -echochange;
	else if (style == STYLE_LEVEL) *inc = 0;
	else /*(style == STYLE_RISING)*/ *inc = echochange;
}

/****************************************************************/
/* generates echoes for a single event */

static struct NoteEvent * __regargs doecho(struct EchoTool *tool,
                                           struct NoteEvent *event)
{
	register struct NoteEvent *new;
	register struct NoteEvent *echolist = NULL;
	struct Tool *echonext;	// where the echoes go
	long time;
	short velocity, duration;
	short count, inc;

	if (tool->flags & BRANCH_ECHOES) echonext = tool->tool.branch;
	else echonext = tool->tool.next;
	if (!echonext) return(NULL);

	time = event->time;
	if (tool->flags & DELETE_SOURCE) event->time = -1;
	else if (tool->flags & BRANCH_SOURCE) {
		if (tool->tool.branch) {
			event->tool = tool->tool.branch;
			event->type |= EVENT_BRANCH;
		}
		else event->time = -1;
	}
	else event->tool = tool->tool.next;

/*
THE COMMENTED STUFF WAS REMOVED TO SUPPORT NOTES WITHOUT DURATIONS!!!!!
	if ((event->time > -1) && IsNoteOn(event) && IsRealTime(event)) {
		event->next = CloneNoteOff(event);
		if (!(event->next)) event->time = -1;
	}
*/
	setup(tool,event,&count,&inc);
	if (!count) { /* tool->repeat is 0 */
		if (event->time < 0) {
			functions->freeevent((struct Event *)event);
			event = NULL;
		}
		return(event);
	}
	/* else */
	duration = event->duration;
	if (IsMonoTouch(event)) velocity = event->value;
	else velocity = event->velocity;
	for ( ; count ; count-- ) {
		new = (struct NoteEvent *)functions->dupeevent((struct Event *)event);
		if (new) {
			time += tool->delay;
			velocity += inc;
			if (velocity < 1) {
				if (event->status == MIDI_NOTEON) velocity = 1;
				else velocity = 0;
			}
			else if (velocity > 127) velocity = 127;
			if (tool->flags & BRANCH_ECHOES) new->type |= EVENT_BRANCH;
			else if (tool->flags & BRANCH_SOURCE) new->type &= ~EVENT_BRANCH;
			new->tool = echonext;
			new->status = event->status;
			new->time = time;
			if (IsMonoTouch(event)) new->value = velocity;
			else new->velocity = velocity;
			new->duration = duration;
			new->next = echolist;
			echolist = new;
/*
			if (IsRealTime(new) && IsNoteOn(new)) {
				if (new = CloneNoteOff(new)) {
					new->next = echolist;
					echolist = new;
				}
				else echolist->time = -1;
			}
*/
		}
	}
	echolist = (struct NoteEvent *)functions->List_Cat((struct ListItem *)event,
	                                                   (struct ListItem *)echolist);
	echolist = (struct NoteEvent *)functions->sorteventlist((struct Event *)echolist);
	for ( event = echolist ; event ; event = echolist ) {
		if (event->time >= 0) break;
		else {
			echolist = event->next;
			functions->freeevent((struct Event *)event);
		}
	}
	return(echolist);
}

/*************************************/
static void __regargs SendSecretMsg(struct Tool *tool,UWORD secret_message)
{
	struct Message *msg;

	if (msg = (struct Message *)functions->myalloc(sizeof(struct Message),0)) {
		msg->mn_Node.ln_Type = NT_MESSAGE;
		msg->mn_ReplyPort = NULL;
		msg->mn_Length = secret_message;
		PutMsg(tool->window->UserPort,msg);
	}
}

/*==========================================================================
* dosysex
*
*  INPUT: pointer to StringEvent
* RETURN: pointer to Event
*
* EFFECT: gets new tool state from sysex message
*=========================================================================*/

static __regargs struct Event *dosysex(struct EchoTool *tool,
                                       struct StringEvent *event)
{
	unsigned char *sxdata;
	UWORD sxflag;

	if (tool->flags & USE_SYSEX) {
		if (sxdata = checksysex(event,TOOL_SYSEX_ID,tool->tool.toolid)) {
			sxflag = tool->flags & (USE_SYSEX | FREE_SYSEX);
			memcpy((void *)&SYSEX_START,(void *)sxdata,SX_LENGTH);
			if (sxflag & FREE_SYSEX) tool->flags |= FREE_SYSEX;
			if (sxflag & USE_SYSEX) tool->flags |= USE_SYSEX;
			if (!InToolPad((struct Tool *)tool)) {
				if (tw) SendSecretMsg(&tool->tool,SECRET_MESSAGE);
				if (tool->flags & FREE_SYSEX) {
					functions->freestring(event->string);
					functions->freeevent((struct Event *)event);
					event = NULL;
				}
			}
		}
	}
	return((struct Event *)event);
}

/******************************************************************/
static struct Event * __stdargs __saveds processlist(struct EchoTool *tool,
                                                     struct NoteEvent *list,
															        long start, long stop)
{
	struct NoteEvent *next;
	struct NoteEvent *echolist; //stuff echoed and echoes
	struct NoteEvent *nonecho;  //stuff not echoed
	char *toolstate; // retain original tool state when processing with sysex

	if (!tool->tool.touched) inittool(tool);
	if ((tool->flags & BRANCH_ECHOES) && (!tool->tool.branch)) {
		return((struct Event *)list);
	}
	if (tool->flags & USE_SYSEX) {
		toolstate = functions->myalloc(SX_LENGTH,MEMF_CLEAR);
		if (toolstate) memcpy(toolstate,(char *)&SYSEX_START,SX_LENGTH);
	}
	else toolstate = NULL;

	echolist = NULL;
	nonecho = NULL;
	for ( ; list && (list->time <= stop) ; list = next ) {
		list->tool = tool->tool.next;
		next = list->next;
		if (IsSysEx((struct Event *)list)) {
			dosysex(tool,(struct StringEvent *)list);
			list->next = nonecho;
			nonecho = list;
		}
		else if ( (((list->status == MIDI_NOTEON) && (tool->flags & NOTES)) ||
	          ((list->status == MIDI_PTOUCH) && (tool->flags & PTOUCH)) ||
		       ((list->status == MIDI_MTOUCH) && (tool->flags & MTOUCH)) ||
		       ((list->status == MIDI_PBEND) && (tool->flags & PBEND)) ||
		       ((list->status == MIDI_CCHANGE) && (tool->flags & CCHG) &&
	           ((list->value==tool->controller)||(tool->controller==121)))) ) {
			list->next = NULL;
			echolist = (struct NoteEvent *)functions->List_Cat(
			                                   (struct ListItem *)echolist,
		                                      (struct ListItem *)doecho(tool,list));
		}
		else {
			list->next = nonecho;
			nonecho = list;
		}
	}
	echolist = (struct NoteEvent *)functions->List_Cat((struct ListItem *)echolist,
		                                                (struct ListItem *)nonecho);
	list = (struct NoteEvent *)functions->sorteventlist((struct Event *)echolist);
	if (toolstate) {
		memcpy((char *)&SYSEX_START,toolstate,SX_LENGTH);
		functions->myfree(toolstate,SX_LENGTH);
	}
	return((struct Event *)list);
}

/****************************************************************/
static UWORD __regargs scalecc(UBYTE val, UWORD  min, UWORD max)
{
	if (val == max) min = max;
#pragma msg 100 ignore push
	else if (val != min)
		min = (UWORD)(IDivU(IMulU(IDivU(IMulU(max-min+1,100),127),val),100) + min);
#pragma msg 100 pop
	return(min);
}

/****************************************************************/
static void docontrolchange(struct EchoTool *tool, struct Event *event)
{
	BYTE i = -1;      /* cinfo[] index */
	UBYTE gotone = 0; /* found entry for this cc# */

	if (event->byte1 < 121) { /* none of the midi resets need apply */
		while ( (i = FindController(&tool->cinfo[0],event->byte1,++i)) > -1 ) {
			gotone=1;
			if (tool->cinfo[i].flags & CIFLAG_DISCARD) event->type |= 0x0F;
			switch (i) {
				case 0 :  // state
					if (SwitchOn(event->byte2)) tool->flags |= TOOL_ON;
					else tool->flags &= ~TOOL_ON;
					break;
				case 1 :  // delay
					tool->delay = scalecc(event->byte2,1,MAX_DELAY);
					break;
				case 2 :  // repeat
					tool->repeat = scalecc(event->byte2,0,MAX_REPEATS);
					break;
				case 3 :  // slope
					if (event->byte2 < 32) tool->echostyle = STYLE_FALLING;
					else if (event->byte2 < 64) tool->echostyle = STYLE_LEVEL;
					else if (event->byte2 < 96) tool->echostyle = STYLE_RISING;
					else tool->echostyle = STYLE_RANDOM;
 					break;
				case 4 :  // branch source
					if (event->byte2 < 43)
						tool->flags = (tool->flags & ~BRANCH_SOURCE) | DELETE_SOURCE;
					else if (event->byte2 < 86)
						tool->flags &= ~(BRANCH_SOURCE | DELETE_SOURCE);
					else tool->flags = (tool->flags & ~DELETE_SOURCE) | BRANCH_SOURCE;
					break;
				case 5 :  // branch echoes
					if (SwitchOn(event->byte2)) tool->flags |= BRANCH_ECHOES;
					else tool->flags &= ~BRANCH_ECHOES;
					break;
				case 6 :  // change
					tool->change = (char)scalecc(event->byte2,0,MAX_CHANGE);
					break;
			}
		}
		if (gotone && tw && !InToolPad(&tool->tool))
			SendSecretMsg(&tool->tool,SECRET_CC);
	}
}


/****************************************************************/
static struct Event * __stdargs __saveds processevent(struct NoteEvent *event)

{
	register struct EchoTool *tool = (struct EchoTool *)event->tool;
	struct NoteEvent *next;
	char eventtype;

	if (!tool->tool.touched) inittool(tool);
	/* do realtime ccs first; can also echo ccs later */
	if (IsCtrlChg(event) && (tool->flags & USE_CONTROLS)) {
		eventtype = event->type; // save this to restore if it should also be echoed
		docontrolchange(tool,(struct Event *)event);
	}

	if (IsSysEx((struct Event *)event)) {
		if (event = (struct NoteEvent *)dosysex(tool,(struct StringEvent *)event))
			event->tool = tool->tool.next;
		return((struct Event *)event);
	}
/* THE COMMENTED STUFF WAS REMOVED TO DEAL WITH NOTES WITH NO DURATIONS !!!!!!!
	else if (event->status == MIDI_NOTEOFF) {
		if ((tool->flags & NOTES) && (tool->flags & TOOL_ON)) event->type |= 0x0F;
	}
*/
	else if ( (tool->flags & TOOL_ON) &&
		       (((event->status == MIDI_NOTEON) && (tool->flags & NOTES)) ||
/**/				 ((event->status == MIDI_NOTEOFF) && (tool->flags & NOTES)) ||
	          ((event->status == MIDI_PTOUCH) && (tool->flags & PTOUCH)) ||
		       ((event->status == MIDI_MTOUCH) && (tool->flags & MTOUCH)) ||
		       ((event->status == MIDI_PBEND) && (tool->flags & PBEND)) ||
		       ((event->status == MIDI_CCHANGE) && (tool->flags & CCHG) &&
	           ((event->value==tool->controller)||(tool->controller==121)))) ) {
		if (IsCtrlChg(event) && ((event->type & 0x0F)==0x0F)) {
			event->type = eventtype; // in case docc() flagged it for delete
		}
		event = doecho(tool,event);
	}
	else { // didn't echo it; now what?
		if (!IsCtrlChg(event)) { // don't dis or branch cc resets, butthole
			if (tool->flags & DELETE_SOURCE) event->type |= 0x0F;
			else if (tool->flags & BRANCH_SOURCE) {
				if (tool->tool.branch) event->tool = tool->tool.branch;
				else event->type |= 0x0F;
			}
			else event->tool = tool->tool.next;
/*
			if ((event->status == MIDI_NOTEON) && ((event->type & 0x0F) != 0x0F)) {
				event->next = CloneNoteOff(event);
				if (!event->next) event->type |= 0x0F;
			}
*/
		}
		else event->tool = tool->tool.next;
	}
	while (event && ((event->type & 0x0F) == 0x0F)) {
		next = event->next;
		functions->freeevent((struct Event *)event);
		event = next;
	}
	return((struct Event *)event);
}

static struct ToolMaster master;

/******************************************************************/
struct ToolMaster * __stdargs __saveds inittoolmaster(void)

{
	memset((char *)&master,0,sizeof(struct ToolMaster));
	master.toolsize = sizeof(struct EchoTool);
	master.toolid = NEco_TOOLID;
	master.image = &IconDown;
	master.upimage = &IconUp;
	strcpy(master.name,WTITLE);
	master.edittool = edittool;
	master.processevent = processevent;
	master.processlist = processlist;
	master.tooltype = TOOL_NORMAL | TOOL_BRANCHOUT;
	return(&master);
}
